/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2010  Rico Rommel

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "sqlresource.h"
#include "configdialog.h"
#include "settings.h"
#include "settingsadaptor.h"

#include <QtDBus/QDBusConnection>
#include <QDebug>
#include <QSqlError>
#include <QMessageBox>
#include <QDialog>
#include <QSqlField>
#include <akonadi/itemfetchscope.h>
#include <akonadi/changerecorder.h>



using namespace Akonadi;

SQLResource::SQLResource( const QString &id )
        : ResourceBase( id )
{
      Online = false;
      
    KWalletFolder = "Akonadi SQL Resource";
    KWalletAccessType = 0;	//0 = do nothing, 1 = read, 2 = write
    wallet = NULL;

    

    

    createAddressTypeMap();
    createPhoneTypeMap();
    createTypeAttributesMap();

    MultipleFields << AddresseeTypes::EMail << AddresseeTypes::Custom << AddresseeTypes::Ignore;
    MultipleFields << PhoneTypeMap.values();

    new SettingsAdaptor( Settings::self() );
    QDBusConnection::sessionBus().registerObject( QLatin1String( "/Settings" ),
            Settings::self(), QDBusConnection::ExportAdaptors );

   changeRecorder()->itemFetchScope().fetchFullPayload();

    
    readSettings();
    setName( DBSettings.resourcename );

    if (DBSettings.savepassword) 
      getDatabaseLoginFromKWallet();
      else
	getDatabaseLoginFromDialog();
    
    

}

SQLResource::~SQLResource()
{
}

void SQLResource::retrieveCollections()
{
    Collection c;
    c.setParent( Collection::root() );
    c.setRemoteId( DBSettings.table );
    c.setName( name() );
    QStringList mimeTypes;
    mimeTypes << "text/directory";
    c.setContentMimeTypes( mimeTypes );
    Collection::List list;
    list << c;
    
 
    collectionsRetrieved( list );

}

void SQLResource::retrieveItems( const Akonadi::Collection &collection )
{
    Q_UNUSED( collection );
    
   
    if (Online) {
    Item::List items;
    QSqlQuery query("SELECT `" + DBSettings.primary + "` from `" + DBSettings.table + "`", db);
    while (query.next()) {
        QString id = DBSettings.database + "_" + DBSettings.table + "_" + query.record().value(DBSettings.primary).toString();
        Item item(QLatin1String( "text/directory"));
        item.setRemoteId(id);
        items  << item;
    }
    itemsRetrieved( items );
    }
    else cancelTask(i18n("Database offline!"));

}

bool SQLResource::retrieveItem( const Akonadi::Item &item, const QSet<QByteArray> &parts )
{
    Q_UNUSED( item );
    Q_UNUSED( parts );

    if (Online) {
    QString id = item.remoteId().remove(DBSettings.database + "_" + DBSettings.table + "_");
    KABC::Addressee addressee = readFromDatabase( id );

    Item newItem( item );
    newItem.setPayload<KABC::Addressee>( addressee );

    itemRetrieved( newItem );

    return true;
    }
    else {
      cancelTask(i18n("Database offline!"));
      return false;
    }
    

}

void SQLResource::aboutToQuit()
{
    db.close();

}

void SQLResource::configure( WId windowId )
{
    Q_UNUSED( windowId );

    ConfigDialog dlg( 0 );
    dlg.setHostname(DBSettings.host);
    dlg.setUsername(DBSettings.user);
    dlg.setPassword(DBSettings.password);
    dlg.setDatabase(DBSettings.database);
    dlg.setPort(DBSettings.port);
    dlg.setMultipleKabcFields(MultipleFields);
    dlg.setTypeAttributesMap(TypeAttributesMap);
    dlg.setTable(DBSettings.table);
    dlg.setMap(DBSettings.map);
    dlg.setDriver(DBSettings.driver);
    dlg.setSavePassword(DBSettings.savepassword);
    dlg.setReadOnly(DBSettings.readonly);
    dlg.setAutogenerate(DBSettings.autogenerate);
    dlg.setResourceName(DBSettings.resourcename);

    if (dlg.exec() == QDialog::Accepted) {
        DBSettings.host = dlg.hostname();
        DBSettings.database = dlg.database();
        DBSettings.driver = dlg.driver();
        DBSettings.password = dlg.password();
        DBSettings.port = dlg.port();
        DBSettings.table = dlg.table();
        DBSettings.user = dlg.username();
        DBSettings.savepassword = dlg.savePassword();
        DBSettings.map = dlg.map();
        DBSettings.autogenerate = dlg.autogenerate();
        DBSettings.readonly = dlg.readOnly();
        DBSettings.resourcename = dlg.resourceName();
        setName( DBSettings.resourcename );
        saveSettings();
        initDB();
        clearCache();
        synchronize();
        emit configurationDialogAccepted();
    }
    else
        emit configurationDialogRejected();


}

void SQLResource::itemAdded( const Akonadi::Item &item, const Akonadi::Collection &collection )
{
    Q_UNUSED( item );
    Q_UNUSED( collection );
    if (!DBSettings.readonly) {
      if (Online) {
        KABC::Addressee addressee;

        QSqlQuery query("INSERT INTO " + DBSettings.table + " () VALUES();", db);
        qDebug() << "item added: " << query.lastError().text();
        query.exec("SELECT LAST_INSERT_ID()");
        query.next();
        QString newid =  query.value(0).toString();

        if (item.hasPayload<KABC::Addressee>()) {
            addressee = item.payload<KABC::Addressee>();
            writeToDatabase(addressee, newid);
        }


        Item newItem( item );
        newItem.setRemoteId( DBSettings.database + "_" + DBSettings.table + "_" +  newid );
        newItem.setPayload<KABC::Addressee>( addressee );

        changeCommitted( newItem );
    }
    else cancelTask(i18n("Database offline!"));
          }

    else cancelTask(i18n("Cannot add: readonly-mode"));
}

void SQLResource::itemChanged( const Akonadi::Item &item, const QSet<QByteArray> &parts )
{
    Q_UNUSED( item );
    Q_UNUSED( parts );

    if (!DBSettings.readonly) {
      if (Online) {
        KABC::Addressee addressee;
        if ( item.hasPayload<KABC::Addressee>() )
            addressee = item.payload<KABC::Addressee>();
        QString id = item.remoteId().remove(DBSettings.database + "_" + DBSettings.table + "_");
        writeToDatabase(addressee, id );

        Item newItem( item );
        newItem.setPayload<KABC::Addressee>( addressee );

        changeCommitted( newItem );
      }
      else cancelTask(i18n("Database offline!"));
    }
    else cancelTask(i18n("Cannot change: readonly-mode"));
}

void SQLResource::itemRemoved( const Akonadi::Item &item )
{
    Q_UNUSED( item );
    if (!DBSettings.readonly) {
        QString id = item.remoteId().remove(DBSettings.database + "_" + DBSettings.table + "_");
        QSqlQuery query("DELETE FROM " + DBSettings.table + " WHERE + " + DBSettings.primary + " = '" + id + "'", db);
        qDebug() << "item " + item.remoteId() + " removed: " << query.lastError().text();


        changeCommitted( item );
    }
    else cancelTask(i18n("Cannot remove: readonly-mode"));
}

void SQLResource::readSettings()
{
    DBSettings.host = Settings::self()->host();
    DBSettings.database = Settings::self()->database();
    DBSettings.port = Settings::self()->port();
    DBSettings.table = Settings::self()->table();
    DBSettings.driver = Settings::self()->driver();
    DBSettings.savepassword = Settings::self()->savepassword();
    DBSettings.autogenerate = Settings::self()->autogenerate();
    DBSettings.readonly = Settings::self()->readonly();
    DBSettings.resourcename = Settings::self()->resourcename();

    QStringList StringMap = Settings::self()->map();
    QStringList pair;
    int n;
    for ( n = 0; n < StringMap.count(); n++) {
        pair = StringMap.at(n).split(":");
        DBSettings.map.insert(pair.at(0),(AddresseeTypes::Types)pair.at(1).toInt());

    }


}

void SQLResource::saveSettings()
{
    Settings::self()->setHost(DBSettings.host);
    Settings::self()->setDatabase(DBSettings.database);
    Settings::self()->setDriver(DBSettings.driver);
    Settings::self()->setPort(DBSettings.port);
    Settings::self()->setTable(DBSettings.table);
    Settings::self()->setSavepassword(DBSettings.savepassword);
    Settings::self()->setAutogenerate(DBSettings.autogenerate);
    Settings::self()->setReadonly(DBSettings.readonly);
    Settings::self()->setResourcename(DBSettings.resourcename);

    QStringList StringMap;

    QMapIterator<QString, AddresseeTypes::Types> i(DBSettings.map);
    while (i.hasNext()) {
        i.next();
        StringMap << i.key() + ":" + QString::number((int)i.value());
    }
    Settings::self()->setMap(StringMap);


    Settings::self()->writeConfig();

    if (DBSettings.savepassword)
      saveDatabaseLoginToKWallet();
}



bool SQLResource::tableExists(QString table)
{
    if (db.tables().indexOf(table) < 0) return false;
    else return true;
}


void SQLResource::initDB()
{
    if (!DBSettings.database.isEmpty()) {
        if (db.isOpen()) db.close();

//        getDataBaseLoginData();

        db = QSqlDatabase::addDatabase(DBSettings.driver, "db");
        db.setHostName(DBSettings.host);
        db.setDatabaseName(DBSettings.database);
        if (db.open(DBSettings.user, DBSettings.password)) {
            QSqlQuery primquery("SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.STATISTICS WHERE INDEX_NAME='PRIMARY' AND INDEX_SCHEMA='" + DBSettings.database + "' AND TABLE_NAME='" + DBSettings.table + "'", db);
            primquery.first () ;
            DBSettings.primary = primquery.value(0).toString();
	    qDebug() << "The database in online!";
	    Online = true;
        }
//        else QMessageBox::Critical();

    }
}

KABC::Address SQLResource::createAddress(KABC::Address::TypeFlag type, QList< AddresseeTypes::Types > list, QSqlRecord record)
{
    QString dbfield;
    KABC::Address address;
    address.setType(type);
    dbfield = DBSettings.map.key(list.at(0));
    if ( !dbfield.isEmpty() )
        address.setCountry(record.value(dbfield).toString());
    dbfield = DBSettings.map.key(list.at(1));
    if ( !dbfield.isEmpty() )
        address.setRegion(record.value(dbfield).toString());
    dbfield = DBSettings.map.key(list.at(2));
    if ( !dbfield.isEmpty() )
        address.setLocality(record.value(dbfield).toString());
    dbfield = DBSettings.map.key(list.at(3));
    if ( !dbfield.isEmpty() )
        address.setStreet(record.value(dbfield).toString());
    dbfield = DBSettings.map.key(list.at(4));
    if ( !dbfield.isEmpty() )
        address.setPostalCode(record.value(dbfield).toString());
    dbfield = DBSettings.map.key(list.at(5));
    if ( !dbfield.isEmpty() )
        address.setCountry(record.value(dbfield).toString());
    return address;

}



void SQLResource::createTypeAttributesMap()
{
    addTypeAttributesToMap(AddresseeTypes::Ignore,  i18n("ignore"), "");
    addTypeAttributesToMap(AddresseeTypes::AdditionalName,  i18n("additional name"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Birthday,  i18n("birthday"), "DATE NOT NULL DEFAULT '1970-00-00'");
    addTypeAttributesToMap(AddresseeTypes::Categories,  i18n("categories"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Custom,  i18n("custom"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Department,  i18n("department"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::FamilyName,  i18n("family name"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::FormattedName,  i18n("formatted name"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::EMail,  i18n("e-mail"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Geo,  i18n("geolocation"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::GivenName,  i18n("given name"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Logo,  i18n("logo"), "BLOB");
    addTypeAttributesToMap(AddresseeTypes::Mailer,  i18n("mailer"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Name,  i18n("identifier"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::NickName,  i18n("nickname"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Note,  i18n("note"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Organization,  i18n("organization"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Photo,  i18n("photo"), "BLOB");
    addTypeAttributesToMap(AddresseeTypes::Prefix,  i18n("prefix"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Role,  i18n("role"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Sound,  i18n("sound"), "BLOB");
    addTypeAttributesToMap(AddresseeTypes::Suffix,  i18n("suffix"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Timezone,  i18n("timezone"), "INT");
    addTypeAttributesToMap(AddresseeTypes::Title,  i18n("title"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::Url,  i18n("url"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CountryDom,  i18n("country (domestic)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::RegionDom,  i18n("region (domestic)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CityDom,  i18n("city (domestic)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostalCodeDom,  i18n("postalcode (domestic)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::StreetDom,  i18n("street (domestic)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostofficeBoxDom,  i18n("postofficebox (domestic)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CountryIntl,  i18n("country (international)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::RegionIntl,  i18n("region (international)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CityIntl,  i18n("city (international)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostalCodeIntl,  i18n("postalcode (international)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::StreetIntl,  i18n("street (international)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostofficeBoxIntl,  i18n("postofficebox (international)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CountryPostal,  i18n("country (postal)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::RegionPostal,  i18n("region (postal)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CityPostal,  i18n("city (postal)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostalCodePostal,  i18n("postalcode (postal)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::StreetPostal,  i18n("street (postal)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostofficeBoxPostal,  i18n("postofficebox (postal)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CountryParcel,  i18n("country (parcel)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::RegionParcel,  i18n("region (parcel)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CityParcel,  i18n("city (parcel)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostalCodeParcel,  i18n("postalcode (parcel)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::StreetParcel,  i18n("street (parcel)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostofficeBoxParcel,  i18n("postofficebox (parcel)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CountryHome,  i18n("country (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::RegionHome,  i18n("region (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CityHome,  i18n("city (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostalCodeHome,  i18n("postalcode (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::StreetHome,  i18n("street (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostofficeBoxHome,  i18n("postofficebox (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CountryWork,  i18n("country (work)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::RegionWork,  i18n("region (work)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::CityWork,  i18n("city (work)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostalCodeWork,  i18n("postalcode (work)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::StreetWork,  i18n("street (work)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PostofficeBoxWork,  i18n("postofficebox (work)"), "TEXT");

    addTypeAttributesToMap(AddresseeTypes::PhoneHome,  i18n("phone (home)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneWork,  i18n("phone (work)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneMsg,  i18n("phone (messager)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneVoice,  i18n("phone (voice)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneFax,  i18n("phone (fax)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneCell,  i18n("phone (mobile)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneVideo,  i18n("phone (video)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneBps,  i18n("phone (bps)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneModem,  i18n("phone (modem)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneCar,  i18n("phone (car)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhoneIsdn,  i18n("phone (isdn)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhonePcs,  i18n("phone (pcs)"), "TEXT");
    addTypeAttributesToMap(AddresseeTypes::PhonePager,  i18n("phone (pager)"), "TEXT");



}

bool SQLResource::createNewField(AddresseeTypes::Types type, QString suffix)
{
    QString fieldname = TypeAttributesMap[type].Text;
    fieldname = makeStringDataBaseCompliant(fieldname);
    return createNewField(fieldname, type, suffix);
}


bool SQLResource::createNewField(QString fieldname, AddresseeTypes::Types type, QString suffix)
{

    if (!suffix.isEmpty())
        fieldname = fieldname + "_" + suffix;

    bool isError;
    //Check if field already exists
    if (!db.record(DBSettings.table).contains(fieldname)) {

        QSqlQuery query("ALTER TABLE " + DBSettings.table + " ADD COLUMN `" + fieldname + "` " + TypeAttributesMap[type].DBType, db);
        qDebug() << "create field: " + fieldname + " error: " + query.lastError().text();
        isError = query.lastError().isValid();
    }
    else {
        isError = false;
        qDebug() << "A field with this name already exists, using it";
    }
    if (!isError) {
        DBSettings.map[fieldname] = type;
        saveSettings();
    }
    return !isError;
}

bool SQLResource::writeValueToDataBase(QString col, QString val, QString id)
{
    QSqlQuery query("UPDATE " + DBSettings.table + " SET `" + col + "`='" + val + "' WHERE `" + DBSettings.primary + "`=" + id, db);
    qDebug() << col + ": write value: " + val + " error: " + query.lastError().text();
    return !query.lastError().isValid();

}

bool SQLResource::writeBlobToDataBase(QString col, QByteArray val, QString id)
{
    QSqlQuery query(db);
    query.prepare("UPDATE " + DBSettings.table + " SET `" + col + "`=:data  WHERE `" + DBSettings.primary + "`=" + id);
    query.bindValue(":data", val);
    query.exec();
    qDebug() << "write blob: " << query.lastError().text();
    return !query.lastError().isValid();
}



void SQLResource::writeToDatabase(KABC::Addressee addressee, QString id)
{

    QByteArray ba;
    QBuffer buffer(&ba);
    int n;
    AddresseeTypes::Types CurrentType;

    CurrentType = AddresseeTypes::AdditionalName;
    if (!addressee.additionalName().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.additionalName(), id);
    }

    CurrentType = AddresseeTypes::Birthday;
    if (!addressee.birthday().isNull() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.birthday().toString(Qt::ISODate), id);
    }

    CurrentType = AddresseeTypes::Categories;
    if (!addressee.categories().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        QStringList CategoriesList = addressee.categories();
        QString CategoriesString;
        for ( n = 0; n < CategoriesList.count(); n++) {
            CategoriesString +=  CategoriesList.at(n);
            if ( n < CategoriesList.count()-1) CategoriesString += ";";
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),CategoriesString, id);
    }


    CurrentType = AddresseeTypes::Department;
    if (!addressee.department().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.department(), id);
    }


    CurrentType = AddresseeTypes::FamilyName;
    if (!addressee.familyName().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.familyName(), id);
    }

    CurrentType = AddresseeTypes::FormattedName;
    if (!addressee.formattedName().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.formattedName(), id);
    }


    CurrentType = AddresseeTypes::Geo;
    if (!addressee.geo().isValid() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        QString geovalue = QString::number(addressee.geo().latitude()) + ";" + QString::number(addressee.geo().longitude());
        writeValueToDataBase(DBSettings.map.key(CurrentType),geovalue, id);
    }

    CurrentType = AddresseeTypes::GivenName;
    if (!addressee.givenName().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.givenName(), id);
    }

    CurrentType = AddresseeTypes::Logo;
    if (!addressee.logo().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty() ) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        buffer.open(QIODevice::WriteOnly);
        addressee.logo().data().save(&buffer, "PNG");
        writeBlobToDataBase(DBSettings.map.key(CurrentType),ba, id);
        buffer.close();
        ba.clear();
    }

    CurrentType = AddresseeTypes::Mailer;
    if (!addressee.mailer().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.mailer(), id);
    }

    CurrentType = AddresseeTypes::Name;
    if (!addressee.name().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.name(), id);
    }

    CurrentType = AddresseeTypes::NickName;
    if (!addressee.nickName().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.nickName(),id );
    }

    CurrentType = AddresseeTypes::Note;
    if (!addressee.note().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.note(), id);
    }

    CurrentType = AddresseeTypes::Organization;
    if (!addressee.organization().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.organization(), id);
    }

    CurrentType = AddresseeTypes::Photo;
    if (!addressee.photo().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty() ) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        buffer.open(QIODevice::WriteOnly);
        addressee.photo().data().save(&buffer, "PNG");
        writeBlobToDataBase(DBSettings.map.key(CurrentType),ba, id);
        buffer.close();
        ba.clear();
    }

    CurrentType = AddresseeTypes::Prefix;
    if (!addressee.prefix().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.prefix(), id);
    }

    CurrentType = AddresseeTypes::Role;
    if (!addressee.role().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.role(), id);
    }

    CurrentType = AddresseeTypes::Sound;
    if (!addressee.sound().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty() ) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeBlobToDataBase(DBSettings.map.key(CurrentType),addressee.sound().data(), id);
        ba.clear();
    }

    CurrentType = AddresseeTypes::Suffix;
    if (!addressee.suffix().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.suffix(), id);
    }

    CurrentType = AddresseeTypes::Timezone;
    if (addressee.timeZone().isValid() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),QString::number(addressee.timeZone().offset()), id);
    }

    CurrentType = AddresseeTypes::Title;
    if (!addressee.title().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.title(), id);
    }

    CurrentType = AddresseeTypes::Url;
    if (!addressee.url().isEmpty() || !DBSettings.map.key(CurrentType).isEmpty()) {
        if (DBSettings.map.key(CurrentType).isEmpty() && DBSettings.autogenerate) {
            createNewField(CurrentType);
        }
        writeValueToDataBase(DBSettings.map.key(CurrentType),addressee.url().url(), id);
    }

    //Ab hier alles ändern: In zugeordnete Felder MUSS geschrieben werden!

    //Mail:
    CurrentType = AddresseeTypes::EMail;
    QStringList maillist = addressee.emails();
//    QStringList mailDBList = DBSettings.map.keys(CurrentType);
    QStringList mailDBList = getOrderedFieldList(CurrentType);
    QString data;
    for ( n = 0; n < mailDBList.count(); n++ ) {
        if ( n < maillist.count() ) data = maillist.at(n);
        else data = "";
        writeValueToDataBase(mailDBList.at(n), data, id);
    }

    // if more mailadresses then fields in table, create new fields
    if (DBSettings.autogenerate) {
        QString newfieldname;
        QString suffix;
        for ( n = mailDBList.count(); n < maillist.count(); n++ ) {
            newfieldname = makeStringDataBaseCompliant( TypeAttributesMap.value(CurrentType).Text );
            suffix = findNextFieldSuffix( newfieldname ) ;
            createNewField(CurrentType, suffix );
            writeValueToDataBase(newfieldname + "_" + suffix, maillist.at(n), id);
        }
    }


    QMapIterator<KABC::PhoneNumber::TypeFlag,  AddresseeTypes::Types> i(PhoneTypeMap);
    while (i.hasNext()) {
        i.next();
        CurrentType = i.value();
        QList<KABC::PhoneNumber> phonelist = addressee.phoneNumbers(i.key());
//        QStringList phoneDBlist = DBSettings.map.keys( CurrentType );
        QStringList phoneDBlist = getOrderedFieldList( CurrentType );

        QString data;

        for ( n = 0; n < phoneDBlist.count(); n++) {
            if ( n < phonelist.count() ) data = phonelist.at( n ).number();

            else data = "";
            writeValueToDataBase(phoneDBlist.at(n), data, id);
        }

        if (DBSettings.autogenerate ) {
            QString newfieldname;
            QString suffix;
            for ( n = phoneDBlist.count(); n < phonelist.count(); n++ ) {
                newfieldname = makeStringDataBaseCompliant( TypeAttributesMap.value( CurrentType ).Text  );
                suffix = findNextFieldSuffix( newfieldname ) ;
                createNewField(CurrentType, suffix );
                writeValueToDataBase(newfieldname + "_" + suffix, phonelist.at(n).number(), id);
            }
        }
    }

    QMapIterator<KABC::Address::TypeFlag, QList <AddresseeTypes::Types> > a(AddressTypeMap);
    while (a.hasNext()) {
        a.next();
        KABC::Address::List addresslist = addressee.addresses(a.key());

        QList< AddresseeTypes::Types> addresseetypeslist = AddressTypeMap.value(a.key());
        if (addresslist.count() > 0 ) {

            KABC::Address address = addresslist.at(0);
            for ( n= 0; n < addresseetypeslist.count(); n++) { // durchlaufen aller Untertypen, AddresseeTypes::CountryDom usw
                CurrentType = addresseetypeslist.at(n);

                switch (n ) {

                case 0:
                    data = address.country();
                    break;
                case  1 :
                    data = address.region();
                    break;
                case 2:
                    data = address.locality();
                    break;
                case 3 :
                    data = address.street();
                    break;
                case 4 :
                    data = address.postalCode();
                    break;
                case 5 :
                    data = address.postOfficeBox();
                    break;

                }
                QString currentfieldname = DBSettings.map.key(CurrentType,"");

                if ( currentfieldname != "" )
                    writeValueToDataBase(currentfieldname,data, id);
                else if ( data != "" && DBSettings.autogenerate ) {
                    createNewField(CurrentType);
                    writeValueToDataBase(DBSettings.map.key(CurrentType),data,id);
                }

            }
        }


    }


    CurrentType = AddresseeTypes::Custom;
    QStringList customlist = addressee.customs();
    QStringList customDBlist = DBSettings.map.keys(CurrentType);
    for ( n = 0; n < customDBlist.count(); n++ ) {
        QString data = addressee.custom("KADDRESSBOOK", customDBlist.at(n));
        writeValueToDataBase(customDBlist.at(n),data,id);

        //remove it from customlist
        int m = 0;
        bool found = true;
        while ( m < customlist.count() && !found) {
            if (customlist.at(m).contains("KADDRESSBOOK-" + customDBlist.at(n)))
                customlist.takeAt(m);
            m++;
        }
    }

    //create new fields
    if (DBSettings.autogenerate) {
        for (n = 0; n < customlist.count(); n++) {
            QString data = customlist.at(n);
            data.remove("KADDRESSBOOK-");
            QStringList data2 = data.split(":");
            createNewField(data2.at(0),AddresseeTypes::Custom);
            writeValueToDataBase(data2.at(0), data2.at(1),id);
        }
    }

}



KABC::Addressee SQLResource::readFromDatabase(QString id)
{
    QSqlQuery query("SELECT * FROM `" + DBSettings.table + "` WHERE `" + DBSettings.primary + "` = '" + id + "'", db);
    KABC::Addressee addressee;
    int n;
    QByteArray ba;

    query.first();

    QString dbfield;
    QStringList dbfieldlist;

    dbfield = DBSettings.map.key(AddresseeTypes::AdditionalName);
    if ( !dbfield.isEmpty() )
        addressee.setAdditionalName(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Birthday);
    if ( !dbfield.isEmpty() )
        addressee.setBirthday(QDateTime::fromString(query.record().value(dbfield).toString(), Qt::ISODate));

    dbfield = DBSettings.map.key(AddresseeTypes::Categories);
    if ( !dbfield.isEmpty() )
        addressee.setCategories(query.record().value(dbfield).toString().split(";"));

    dbfield = DBSettings.map.key(AddresseeTypes::Department);
    if ( !dbfield.isEmpty() )
        addressee.setDepartment(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::FamilyName);
    if ( !dbfield.isEmpty() )
        addressee.setFamilyName(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::FormattedName);
    if ( !dbfield.isEmpty() )
        addressee.setFormattedName(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Geo);
    if ( !dbfield.isEmpty() ) {
        QStringList koorlist = query.record().value(dbfield).toString().split(";");
        if (koorlist.count() == 2) {
            KABC::Geo geo;
            geo.setLatitude(koorlist.at(0).toFloat());
            geo.setLongitude(koorlist.at(1).toFloat());
            addressee.setGeo(geo);
        }
    }

    dbfield = DBSettings.map.key(AddresseeTypes::GivenName);
    if ( !dbfield.isEmpty() )
        addressee.setGivenName(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Logo);
    if ( !dbfield.isEmpty() ) {
        ba = query.record().value(dbfield).toByteArray();
        KABC::Picture logo(QImage::fromData(ba));
        addressee.setLogo(logo);
    }

    dbfield = DBSettings.map.key(AddresseeTypes::Mailer);
    if ( !dbfield.isEmpty() )
        addressee.setMailer(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Name);
    if ( !dbfield.isEmpty() )
        addressee.setName(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::NickName);
    if ( !dbfield.isEmpty() )
        addressee.setNickName(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Note);
    if ( !dbfield.isEmpty() )
        addressee.setNote(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Organization);
    if ( !dbfield.isEmpty() )
        addressee.setOrganization(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Photo);
    if (!dbfield.isEmpty()) {
        ba = query.record().value(dbfield).toByteArray();
        KABC::Picture photo(QImage::fromData(ba));
        addressee.setPhoto(photo);
    }

    dbfield = DBSettings.map.key(AddresseeTypes::Prefix);
    if ( !dbfield.isEmpty() )
        addressee.setPrefix(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Role);
    if ( !dbfield.isEmpty() )
        addressee.setRole(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Sound);
    if ( !dbfield.isEmpty() ) {
        ba = query.record().value(dbfield).toByteArray();
        KABC::Sound sound(ba);
        addressee.setSound(sound);
    }

    dbfield = DBSettings.map.key(AddresseeTypes::Suffix);
    if ( !dbfield.isEmpty() )
        addressee.setSuffix(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Timezone);
    if ( !dbfield.isEmpty() )
        addressee.setTimeZone(KABC::TimeZone(query.record().value(dbfield).toInt()));

    dbfield = DBSettings.map.key(AddresseeTypes::Title);
    if ( !dbfield.isEmpty() )
        addressee.setTitle(query.record().value(dbfield).toString());

    dbfield = DBSettings.map.key(AddresseeTypes::Url);
    if ( !dbfield.isEmpty() )
        addressee.setUrl(query.record().value(dbfield).toString());


    QMapIterator<KABC::Address::TypeFlag, QList<AddresseeTypes::Types> > i(AddressTypeMap);
    while (i.hasNext() ) {
        i.next();
        addressee.insertAddress(createAddress(i.key(), i.value(), query.record()));
    }


    dbfieldlist = DBSettings.map.keys(AddresseeTypes::EMail);
    bool preferred = true;
    for ( n = 0; n < dbfieldlist.count(); n++) {
        if (n == 1) preferred = false;
        addressee.insertEmail(query.record().value(dbfieldlist.at(n)).toString(), preferred);
    }


    QMapIterator<KABC::PhoneNumber::TypeFlag, AddresseeTypes::Types> p(PhoneTypeMap);
    while ( p.hasNext() ) {
        p.next();
        QStringList phonenumberlist = DBSettings.map.keys(p.value());
        for (n = 0; n < phonenumberlist.count(); n++) {
            KABC::PhoneNumber number( query.record().value(phonenumberlist.at(n)).toString(), p.key());
            addressee.insertPhoneNumber(number);

        }
    }

    dbfieldlist = DBSettings.map.keys(AddresseeTypes::Custom);
    for ( n  = 0; n < dbfieldlist.count(); n++) {
        addressee.insertCustom("KADDRESSBOOK", dbfieldlist.at(n), query.record().value(dbfieldlist.at(n)).toString());
    }


    return addressee;
}

void SQLResource::addTypeAttributesToMap(AddresseeTypes::Types type, QString text, QString dbtype)
{
    typeattributes attr;
    attr.Text = text;
    attr.DBType = dbtype;
    TypeAttributesMap.insert(type,attr);
}

QString SQLResource::makeStringDataBaseCompliant(QString name)
{
    name.replace(" ","_");
    name.remove("(");
    name.remove(")");
    return name;
}

QString SQLResource::findNextFieldSuffix(QString name)
{
    QSqlRecord rec = db.record(DBSettings.table);
    int n;
    QString suffix;
    n = -1;
    do {
        n++;
        suffix = QString::number(n);

    } while ( rec.contains(name + "_" + suffix) );

    return suffix;
}

QStringList SQLResource::getOrderedFieldList(AddresseeTypes::Types type)
{
    QStringList list;
    QSqlRecord rec = db.record(DBSettings.table);
    int n;
    for ( n = 0; n < rec.count(); n++ ) {
        QString field = rec.field(n).name();
        if ( DBSettings.map.value(field) == type ) list << field;
    }
    return list;
}


void SQLResource::createAddressTypeMap()
{
QList <AddresseeTypes::Types> DomList, IntlList, PostalList, ParcelList, HomeList, WorkList;
    DomList << AddresseeTypes::CountryDom << AddresseeTypes::RegionDom << AddresseeTypes::CityDom << AddresseeTypes::StreetDom << AddresseeTypes::PostalCodeDom << AddresseeTypes::PostofficeBoxDom;
    IntlList << AddresseeTypes::CountryIntl << AddresseeTypes::RegionIntl << AddresseeTypes::CityIntl << AddresseeTypes::StreetIntl << AddresseeTypes::PostalCodeIntl << AddresseeTypes::PostofficeBoxIntl;
    PostalList << AddresseeTypes::CountryPostal << AddresseeTypes::RegionPostal << AddresseeTypes::CityPostal << AddresseeTypes::StreetPostal << AddresseeTypes::PostalCodePostal << AddresseeTypes::PostofficeBoxPostal;
    ParcelList << AddresseeTypes::CountryParcel << AddresseeTypes::RegionParcel << AddresseeTypes::CityParcel << AddresseeTypes::StreetParcel << AddresseeTypes::PostalCodeParcel << AddresseeTypes::PostofficeBoxParcel;
    HomeList << AddresseeTypes::CountryHome << AddresseeTypes::RegionHome << AddresseeTypes::CityHome << AddresseeTypes::StreetHome << AddresseeTypes::PostalCodeHome << AddresseeTypes::PostofficeBoxHome;
    WorkList << AddresseeTypes::CountryWork << AddresseeTypes::RegionWork << AddresseeTypes::CityWork << AddresseeTypes::StreetWork << AddresseeTypes::PostalCodeWork << AddresseeTypes::PostofficeBoxWork;

    AddressTypeMap[KABC::Address::Dom] = DomList;
    AddressTypeMap[KABC::Address::Intl] = IntlList;
    AddressTypeMap[KABC::Address::Postal] = PostalList;
    AddressTypeMap[KABC::Address::Parcel] = ParcelList;
    AddressTypeMap[KABC::Address::Home] = HomeList;
    AddressTypeMap[KABC::Address::Work] = WorkList;
  
}

void SQLResource::createPhoneTypeMap()
{
PhoneTypeMap[KABC::PhoneNumber::Home] = AddresseeTypes::PhoneHome;
    PhoneTypeMap[KABC::PhoneNumber::Work] = AddresseeTypes::PhoneWork;
    PhoneTypeMap[KABC::PhoneNumber::Msg] = AddresseeTypes::PhoneMsg;
    PhoneTypeMap[KABC::PhoneNumber::Voice] = AddresseeTypes::PhoneVoice;
    PhoneTypeMap[KABC::PhoneNumber::Fax] = AddresseeTypes::PhoneFax;
    PhoneTypeMap[KABC::PhoneNumber::Cell] = AddresseeTypes::PhoneCell;
    PhoneTypeMap[KABC::PhoneNumber::Video] = AddresseeTypes::PhoneVideo;
    PhoneTypeMap[KABC::PhoneNumber::Bbs] = AddresseeTypes::PhoneBps;
    PhoneTypeMap[KABC::PhoneNumber::Modem] = AddresseeTypes::PhoneModem;
    PhoneTypeMap[KABC::PhoneNumber::Car] = AddresseeTypes::PhoneCar;
    PhoneTypeMap[KABC::PhoneNumber::Isdn] = AddresseeTypes::PhoneIsdn;
    PhoneTypeMap[KABC::PhoneNumber::Pcs] = AddresseeTypes::PhonePcs;
    PhoneTypeMap[KABC::PhoneNumber::Pager] = AddresseeTypes::PhonePager;
}


void SQLResource::openKWallet()
{
        if (wallet == NULL || !wallet->isOpen()) {
	wallet = KWallet::Wallet::openWallet(KWallet::Wallet::NetworkWallet(), 0, KWallet::Wallet::Asynchronous);

	connect(wallet, SIGNAL(walletOpened(bool)), this, SLOT(accessKWallet(bool)));
	}
	if (wallet->isOpen()) {
	  accessKWallet(true);
	}
}


void SQLResource::getDatabaseLoginFromKWallet()
{
      KWalletAccessType = 1;
      openKWallet();

}

void SQLResource::saveDatabaseLoginToKWallet()
{
  KWalletAccessType = 2;
  openKWallet();

}


void SQLResource::accessKWallet(bool ok)
{
  qDebug() << "access kwallet" << KWalletAccessType;
  int error = -1;

  if (ok) {
   wallet->setFolder(KWalletFolder);

  QMap<QString, QString> MapBuffer;
  switch (KWalletAccessType) {
    case 1 : {
      error = wallet->readMap(DBSettings.resourcename, MapBuffer);
      DBSettings.user = MapBuffer["user"];
      DBSettings.password = MapBuffer["password"];
      break;
          }
    case 2 : {
      if (!wallet->hasFolder(KWalletFolder)) {
	wallet->createFolder(KWalletFolder);
	wallet->setFolder(KWalletFolder);
      }
      MapBuffer["user"] = DBSettings.user;
      MapBuffer["password"] = DBSettings.password;
      error = wallet->writeMap(DBSettings.resourcename,MapBuffer);
      break;
      
    }
    default : break;
  }
  if (KWalletAccessType == 1 ) {
    if (!DBSettings.user.isEmpty())
      gotLoginData();
    else 
        getDatabaseLoginFromDialog();

  }
    KWalletAccessType = 0;

}
}


void SQLResource::getDatabaseLoginFromDialog()
{
  LoginDlg = new LoginDialog(DBSettings.database, DBSettings.host, 0);
  LoginDlg->show();
  connect(LoginDlg, SIGNAL(accepted()), this, SLOT(readLoginDataFromDialog()));
}

void SQLResource::readLoginDataFromDialog()
{
  DBSettings.user = LoginDlg->user();
  DBSettings.password = LoginDlg->password();
  DBSettings.savepassword = LoginDlg->savePassword();
  saveSettings();
  gotLoginData();
  
}

void SQLResource::gotLoginData()
{
   qDebug() << "We got the login data!";
  initDB();
}






AKONADI_RESOURCE_MAIN( SQLResource )

#include "sqlresource.moc"
